@@ -1,18 +1,21 @@
 import os
 import re
 import sys
-import pip as pip_mod
 import time
-import shutil
+import select
 import pkgutil
 import logging
 import warnings
 import threading
+import traceback
+import pip as pip_mod
+import shutil
 try:
     import subprocess32 as subprocess
 except Exception:
     import subprocess
 import six
+from datetime import datetime
 from localstack import constants, config
 
 # set up logger
@@ -21,6 +24,9 @@ LOG = logging.getLogger(os.path.basename(__file__))
 # maps plugin scope ("services", "commands") to flags which indicate whether plugins have been loaded
 PLUGINS_LOADED = {}
 
+# predefined list of plugin modules, to speed up the plugin loading at startup
+PLUGIN_MODULES = ['localstack', 'localstack_ext']
+
 # marker for extended/ignored libs in requirements.txt
 IGNORED_LIB_MARKER = '#extended-lib'
 BASIC_LIB_MARKER = '#basic-lib'
@@ -39,14 +45,25 @@ PLUGIN_SCOPE_COMMANDS = 'commands'
 # maps from API names to list of other API names that they depend on
 API_DEPENDENCIES = {
     'dynamodbstreams': ['kinesis'],
-    'lambda': ['logs'],
-    'es': ['elasticsearch']
+    'es': ['elasticsearch'],
+    'lambda': ['logs']
 }
 # composites define an abstract name like "serverless" that maps to a set of services
 API_COMPOSITES = {
-    'serverless': ['iam', 'lambda', 'dynamodb', 'apigateway', 's3', 'sns']
+    'serverless': ['cloudformation', 'iam', 'sts', 'lambda', 'dynamodb', 'apigateway', 's3'],
+    'cognito': ['cognito-idp', 'cognito-identity']
 }
 
+# name of main Docker container
+MAIN_CONTAINER_NAME = 'localstack_main'
+
+# environment variable that indicates that we're executing in
+# the context of the script that starts the Docker container
+ENV_SCRIPT_STARTING_DOCKER = 'LS_SCRIPT_STARTING_DOCKER'
+
+# semaphore for locking access to Popen
+mutex_popen = threading.Semaphore(1)
+
 
 def bootstrap_installation():
     try:
@@ -63,7 +80,7 @@ def install_dependencies():
     reqs_copy_file = os.path.join(root_folder, 'localstack', 'requirements.copy.txt')
     if not os.path.exists(reqs_copy_file):
         shutil.copy(reqs_file, reqs_copy_file)
-    with open(reqs_file) as f:
+    with open(reqs_copy_file) as f:
         requirements = f.read()
     install_requires = []
     for line in re.split('\n', requirements):
@@ -74,11 +91,17 @@ def install_dependencies():
     LOG.info('Lazily installing missing pip dependencies, this could take a while: %s' %
              ', '.join(install_requires))
     args = ['install'] + install_requires
+    return run_pip_main(args)
+
+
+def run_pip_main(args):
     if hasattr(pip_mod, 'main'):
-        pip_mod.main(args)
-    else:
-        import pip._internal
-        pip._internal.main(args)
+        return pip_mod.main(args)
+    import pip._internal
+    if hasattr(pip._internal, 'main'):
+        return pip._internal.main(args)
+    import pip._internal.main
+    return pip._internal.main.main(args)
 
 
 def load_plugin_from_path(file_path, scope=None):
@@ -93,13 +116,15 @@ def load_plugin_from_path(file_path, scope=None):
             exec('from %s.plugins import %s' % (module, method_name), namespace)
             method_to_execute = namespace[method_name]
         except Exception as e:
-            if not re.match(r'.*cannot import name .*%s.*' % method_name, str(e)):
+            if (not re.match(r'.*cannot import name .*%s.*' % method_name, str(e)) and
+                    ('No module named' not in str(e))):
                 LOG.debug('Unable to load plugins from module %s: %s' % (module, e))
             return
         try:
             return method_to_execute()
         except Exception as e:
-            LOG.warning('Unable to load plugins from file %s: %s' % (file_path, e))
+            if not os.environ.get(ENV_SCRIPT_STARTING_DOCKER):
+                LOG.warning('Unable to load plugins from file %s: %s' % (file_path, e))
 
 
 def load_plugins(scope=None):
@@ -107,17 +132,30 @@ def load_plugins(scope=None):
     if PLUGINS_LOADED.get(scope):
         return PLUGINS_LOADED[scope]
 
+    t1 = now_utc()
     setup_logging()
 
     loaded_files = []
     result = []
-    for module in pkgutil.iter_modules():
+
+    # Use a predefined list of plugin modules for now, to speed up the plugin loading at startup
+    # search_modules = pkgutil.iter_modules()
+    search_modules = PLUGIN_MODULES
+
+    for module in search_modules:
         file_path = None
-        if six.PY3 and not isinstance(module, tuple):
-            file_path = '%s/%s/plugins.py' % (module.module_finder.path, module.name)
+        if isinstance(module, six.string_types):
+            loader = pkgutil.get_loader(module)
+            if loader:
+                path = getattr(loader, 'path', '') or getattr(loader, 'filename', '')
+                if '__init__.py' in path:
+                    path = os.path.dirname(path)
+                file_path = os.path.join(path, 'plugins.py')
+        elif six.PY3 and not isinstance(module, tuple):
+            file_path = os.path.join(module.module_finder.path, module.name, 'plugins.py')
         elif six.PY3 or isinstance(module[0], pkgutil.ImpImporter):
             if hasattr(module[0], 'path'):
-                file_path = '%s/%s/plugins.py' % (module[0].path, module[1])
+                file_path = os.path.join(module[0].path, module[1], 'plugins.py')
         if file_path and file_path not in loaded_files:
             plugin_config = load_plugin_from_path(file_path, scope=scope)
             if plugin_config:
@@ -125,15 +163,26 @@ def load_plugins(scope=None):
             loaded_files.append(file_path)
     # set global flag
     PLUGINS_LOADED[scope] = result
+
+    # debug plugin loading time
+    load_time = now_utc() - t1
+    if load_time > 5:
+        LOG.debug('Plugin loading took %s sec' % load_time)
+
     return result
 
 
 def docker_container_running(container_name):
-    output = to_str(run("docker ps --format '{{.Names}}'"))
-    container_names = re.split(r'\s+', output.replace('\n', ' '))
+    container_names = get_docker_container_names()
     return container_name in container_names
 
 
+def get_docker_container_names():
+    output = to_str(run("docker ps --format '{{.Names}}'"))
+    container_names = re.split(r'\s+', output.strip().replace('\n', ' '))
+    return container_names
+
+
 def setup_logging():
     # determine and set log level
     log_level = logging.DEBUG if is_debug() else logging.INFO
@@ -225,15 +274,15 @@ def start_infra_locally():
 
 def start_infra_in_docker():
 
-    container_name = 'localstack_main'
+    container_name = MAIN_CONTAINER_NAME
 
     if docker_container_running(container_name):
         raise Exception('LocalStack container named "%s" is already running' % container_name)
 
+    os.environ[ENV_SCRIPT_STARTING_DOCKER] = '1'
+
     # load plugins before starting the docker container
     plugin_configs = load_plugins()
-    plugin_run_params = ' '.join([
-        entry.get('docker', {}).get('run_flags', '') for entry in plugin_configs])
 
     # prepare APIs
     canonicalize_api_names()
@@ -246,6 +295,22 @@ def start_infra_in_docker():
     service_ports = config.SERVICE_PORTS
     force_noninteractive = os.environ.get('FORCE_NONINTERACTIVE', '')
 
+    # get run params
+    plugin_run_params = ' '.join([
+        entry.get('docker', {}).get('run_flags', '') for entry in plugin_configs])
+
+    # get port ranges defined via DOCKER_FLAGS (if any)
+    regex = r'.*-p\s+([0-9]+)(\-([0-9]+))?:([0-9]+)(\-[0-9]+)?.*'
+    match = re.match(regex, user_flags)
+    start = end = 0
+    if match:
+        start = int(match.group(1))
+        end = int(match.group(3) or match.group(1))
+
+    def is_mapped(start_port, end_port=None):
+        existing_range = range(start, end)
+        return (int(start_port) in existing_range) or (start_port and int(start_port) in existing_range)
+
     # construct port mappings
     ports_list = sorted(service_ports.values())
     start_port = 0
@@ -262,15 +327,16 @@ def start_infra_in_docker():
         elif i >= len(ports_list) - 1:
             port_ranges.append([start_port, ports_list[i]])
         last_port = ports_list[i]
-    port_mappings = ' '.join(
+    port_mappings = ' '.join([
         '-p {start}-{end}:{start}-{end}'.format(start=entry[0], end=entry[1])
         if entry[0] < entry[1] else '-p {port}:{port}'.format(port=entry[0])
-        for entry in port_ranges)
+        for entry in port_ranges if not is_mapped(entry[0], entry[1])])
 
     if services:
         port_mappings = ''
-        for service, port in service_ports.items():
-            port_mappings += ' -p {port}:{port}'.format(port=port)
+        for port in set(service_ports.values()):
+            if not is_mapped(port):
+                port_mappings += ' -p {port}:{port}'.format(port=port)
 
     env_str = ''
     for env_var in config.CONFIG_ENV_VARS:
@@ -282,7 +348,7 @@ def start_infra_in_docker():
     data_dir = os.environ.get('DATA_DIR', None)
     if data_dir is not None:
         container_data_dir = '/tmp/localstack_data'
-        data_dir_mount = '-v "%s:%s" ' % (data_dir, container_data_dir)
+        data_dir_mount = '-v "%s:%s"' % (data_dir, container_data_dir)
         env_str += '-e DATA_DIR="%s" ' % container_data_dir
 
     interactive = '' if force_noninteractive or in_ci() else '-it '
@@ -291,19 +357,21 @@ def start_infra_in_docker():
     user_flags = '%s ' % user_flags if user_flags else user_flags
     entrypoint = '%s ' % entrypoint if entrypoint else entrypoint
     plugin_run_params = '%s ' % plugin_run_params if plugin_run_params else plugin_run_params
-
-    container_name = 'localstack_main'
+    web_ui_flags = ''
+    if config.START_WEB:
+        web_ui_flags = '-p {p}:{p} -p {p1}:{p1} '.format(p=config.PORT_WEB_UI, p1=config.PORT_WEB_UI_SSL)
 
     docker_cmd = ('%s run %s%s%s%s%s' +
         '--rm --privileged ' +
         '--name %s ' +
-        '-p 8080:8080 %s %s' +
+        '%s %s %s ' +
         '-v "%s:/tmp/localstack" -v "%s:%s" ' +
         '-e DOCKER_HOST="unix://%s" ' +
         '-e HOST_TMP_FOLDER="%s" "%s" %s') % (
             config.DOCKER_CMD, interactive, entrypoint, env_str, user_flags, plugin_run_params,
-            container_name, port_mappings, data_dir_mount, config.TMP_FOLDER, config.DOCKER_SOCK,
-            config.DOCKER_SOCK, config.DOCKER_SOCK, config.HOST_TMP_FOLDER, image_name, cmd
+            container_name, web_ui_flags, port_mappings, data_dir_mount,
+            config.TMP_FOLDER, config.DOCKER_SOCK, config.DOCKER_SOCK, config.DOCKER_SOCK,
+            config.HOST_TMP_FOLDER, image_name, cmd
     )
 
     mkdir(config.TMP_FOLDER)
@@ -342,6 +410,10 @@ def start_infra_in_docker():
 # UTIL FUNCTIONS
 # ---------------
 
+def now_utc():
+    epoch = datetime.utcfromtimestamp(0)
+    return (datetime.utcnow() - epoch).total_seconds()
+
 
 def to_str(obj, errors='strict'):
     return obj.decode('utf-8', errors) if isinstance(obj, six.binary_type) else obj
@@ -355,10 +427,114 @@ def in_ci():
     return False
 
 
-def run(cmd, asynchronous=False):
-    if asynchronous:
-        return subprocess.Popen(cmd, shell=True)
-    return subprocess.check_output(cmd, shell=True)
+class FuncThread(threading.Thread):
+    """ Helper class to run a Python function in a background thread. """
+
+    def __init__(self, func, params=None, quiet=False):
+        threading.Thread.__init__(self)
+        self.daemon = True
+        self.params = params
+        self.func = func
+        self.quiet = quiet
+
+    def run(self):
+        try:
+            self.func(self.params)
+        except Exception:
+            if not self.quiet:
+                LOG.warning('Thread run method %s(%s) failed: %s' %
+                    (self.func, self.params, traceback.format_exc()))
+
+    def stop(self, quiet=False):
+        if not quiet and not self.quiet:
+            LOG.warning('Not implemented: FuncThread.stop(..)')
+
+
+def run(cmd, print_error=True, asynchronous=False, stdin=False,
+        stderr=subprocess.STDOUT, outfile=None, env_vars=None, inherit_cwd=False,
+        inherit_env=True, tty=False):
+    # don't use subprocess module in Python 2 as it is not thread-safe
+    # http://stackoverflow.com/questions/21194380/is-subprocess-popen-not-thread-safe
+    # TODO: should be removed, now that Python 2 has reached its EOL
+    if six.PY2:
+        import subprocess32 as subprocess
+    else:
+        import subprocess
+
+    env_dict = os.environ.copy() if inherit_env else {}
+    if env_vars:
+        env_dict.update(env_vars)
+    env_dict = dict([(k, to_str(str(v))) for k, v in env_dict.items()])
+
+    if tty:
+        asynchronous = True
+        stdin = True
+
+    try:
+        cwd = os.getcwd() if inherit_cwd else None
+        if not asynchronous:
+            if stdin:
+                return subprocess.check_output(cmd, shell=True, stderr=stderr, env=env_dict,
+                    stdin=subprocess.PIPE, cwd=cwd)
+            output = subprocess.check_output(cmd, shell=True, stderr=stderr, env=env_dict, cwd=cwd)
+            return output.decode(config.DEFAULT_ENCODING)
+
+        # subprocess.Popen is not thread-safe, hence use a mutex here.. (TODO: mutex still needed?)
+        with mutex_popen:
+            stdin_arg = subprocess.PIPE if stdin else None
+            stdout_arg = open(outfile, 'ab') if isinstance(outfile, six.string_types) else outfile
+            stderr_arg = stderr
+            if tty:
+                # Note: leave the "pty" import here (not supported in Windows)
+                import pty
+                master_fd, slave_fd = pty.openpty()
+                stdin_arg = slave_fd
+                stdout_arg = stderr_arg = None
+
+            # start the actual sub process
+            kwargs = {}
+            if is_linux() or is_mac_os():
+                kwargs['preexec_fn'] = os.setsid
+            process = subprocess.Popen(cmd, shell=True, stdin=stdin_arg, bufsize=-1,
+                stderr=stderr_arg, stdout=stdout_arg, env=env_dict, cwd=cwd, **kwargs)
+
+            if tty:
+                # based on: https://stackoverflow.com/questions/41542960
+                def pipe_streams(*args):
+                    while process.poll() is None:
+                        r, w, e = select.select([sys.stdin, master_fd], [], [])
+                        if sys.stdin in r:
+                            d = os.read(sys.stdin.fileno(), 10240)
+                            os.write(master_fd, d)
+                        elif master_fd in r:
+                            o = os.read(master_fd, 10240)
+                            if o:
+                                os.write(sys.stdout.fileno(), o)
+
+                FuncThread(pipe_streams).start()
+
+            return process
+    except subprocess.CalledProcessError as e:
+        if print_error:
+            print("ERROR: '%s': exit code %s; output: %s" % (cmd, e.returncode, e.output))
+            sys.stdout.flush()
+        raise e
+
+
+def is_mac_os():
+    try:
+        out = to_str(subprocess.check_output('uname -a', shell=True))
+        return 'Darwin' in out
+    except subprocess.CalledProcessError:
+        return False
+
+
+def is_linux():
+    try:
+        out = to_str(subprocess.check_output('uname -a', shell=True))
+        return 'Linux' in out
+    except subprocess.CalledProcessError:
+        return False
 
 
 def mkdir(folder):
